/*  Copyright (C) 2009 Thilo Molitor
    Based on some code I found somewhere in the internet.
    Sorry, but I don't remember where I found it, if anyone knows please send me an email.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>


//LED Pins
#define DDRLED1			DDRD
#define DDRLED2			DDRD
#define PINLED1			PD5
#define PINLED2			PD6
#define PORTLED1		PORTD
#define PORTLED2		PORTD

//Bootloaderbefehle von protokoll v0.1
#define CMD_CONNECT		4
#define CMD_ERASE_FLASH 	5
#define CMD_WRITE_FLASH 	6
#define CMD_READ_FLASH		7
#define CMD_FLASH_BLOCKSIZE 	8
#define CMD_JUMP_PROGRAM 	9
//Bootloaderbefehle von protokoll v0.2
#define CMD_READ_LFUSE 		10
#define CMD_READ_HFUSE 		11
#define CMD_READ_EEPROM		12
#define CMD_WRITE_EEPROM	13
//Bootloaderbefehle von protokoll v0.3
#define CMD_EEPROM_BLOCKSIZE	14

//Messages von protokoll v0.1
#define MSG_END			255
#define MSG_OVERWRITE_ERROR	4
#define MSG_CHKSUM_ERROR	5
#define MSG_PROTOKOLL_H		0
#define MSG_PROTOKOLL_L		3
//keine neuen messages von protokoll v0.2
//keine neuen messages von protokoll v0.3


//adjust uart baudrate in makefile
//adjust BOOTLOADER_ADDRESS in Makefile if you want to change the size of the bootloader)


//#define recchar_safe() {int }

#define	UART_BAUD_RATE_LOW	 UBRRL
#define	UART_STATUS_REG	         UCSRA
#define	UART_CONTROL_REG	 UCSRB
#define	UART_ENABLE_TRANSMITTER	 TXEN
#define	UART_ENABLE_RECEIVER	 RXEN
#define	UART_TRANSMIT_COMPLETE	 TXC
#define	UART_RECEIVE_COMPLETE	 RXC
#define	UART_DATA_REG	         UDR
#define UART_DOUBLE_SPEED        U2X

/*
 * Macro to calculate UBBR from XTAL and baudrate
 */
#define UART_BAUD_SELECT(baudRate,xtalCpu) (((float)(xtalCpu))/(((float)(baudRate))*16.0)-1.0+0.5)


typedef uint16_t address_t;


/*
 * since this bootloader is not linked against the avr-gcc crt1 functions,
 * to reduce the code size, we need to provide our own initialization
 */
void __jumpMain     (void) __attribute__ ((naked)) __attribute__ ((section (".init9")));

void __jumpMain(void)
{    
    asm volatile ( ".set __stack, %0" :: "i" (RAMEND) );
    asm volatile ( "clr __zero_reg__" );                       // GCC depends on register r1 set to 0
    asm volatile ( "out %0, __zero_reg__" :: "I" (_SFR_IO_ADDR(SREG)) );  // set SREG to 0
    asm volatile ( "rjmp main");                               // jump to main()
}


/*
 * send single byte to USART, wait until transmission is completed
 */
static void sendchar(char c)
{
    UART_DATA_REG = c;                                   // prepare transmission
    while(!(UART_STATUS_REG & (1 << UART_TRANSMIT_COMPLETE)));// wait until byte sent
    UART_STATUS_REG |= (1 << UART_TRANSMIT_COMPLETE);          // delete TXCflag
}

/*
 * Read single byte from USART, block if no data available
 */
static unsigned char recchar(void)
{
    while(!(UART_STATUS_REG & (1 << UART_RECEIVE_COMPLETE)));  // wait for data
    return UART_DATA_REG;
}

/*
 * Read single byte from USART, block for time ms if no data available
 */
static int recchar_noblock(unsigned int time)
{
	unsigned int counter=0;
	unsigned int maxcounter=time/10;
	while((!(UART_STATUS_REG & (1 << UART_RECEIVE_COMPLETE)))&&(counter<maxcounter))
	{
		_delay_ms(10);
		counter++;
	}// wait for data
	if(UART_STATUS_REG & (1 << UART_RECEIVE_COMPLETE))
		return UART_DATA_REG;
	else
		return -1;
}


int main(void)
{
	unsigned char 	c;
	unsigned char 	message[520];
	unsigned int  	size=0, msgpos=0, i=0;
	unsigned char 	chksum=0;
	uint32_t	page=0;
	char stop=1;
	
	
	/* PROG_PIN pulled low, indicate with LED that bootloader is active */
	DDRLED1  |= (1<<PINLED1);
	DDRLED2  |= (1<<PINLED2);
	PORTLED1 |= (1<<PINLED1);
	/*
	 * Init UART
	 * set baudrate and enable USART receiver and transmiter without interrupts 
	 */
	UART_BAUD_RATE_LOW = UART_BAUD_SELECT(BAUDRATE,F_CPU);
	UART_CONTROL_REG   = (1 << UART_ENABLE_RECEIVER) | (1 << UART_ENABLE_TRANSMITTER); 
	
	//2sec auf connection warten
	if(recchar_noblock(2000)==CMD_CONNECT)
	{
		sendchar(MSG_PROTOKOLL_H);
		sendchar(MSG_PROTOKOLL_L);
		sendchar(MSG_END);
		
		//main loop
		msgpos=0;
		for(;stop;)
		{
			waitforcommand:
			c=recchar();
			switch(c)
			{
				case CMD_CONNECT:				//Command connect
					sendchar(MSG_PROTOKOLL_H);
					sendchar(MSG_PROTOKOLL_L);
					sendchar(MSG_END);
					break;
				case CMD_ERASE_FLASH:				//Command erase_flash
					for(page=0; page<APP_END; page+=SPM_PAGESIZE)
						boot_page_erase_safe(page);
					sendchar(MSG_END);
					break;
				case CMD_WRITE_FLASH:				//Command write_flash
					page=(recchar()<<8)|recchar();
					size=(recchar()<<8)|recchar();
					if(size>SPM_PAGESIZE)
					{
						//Fehler: wir wollen uns nicht selbst berschreiben
						sendchar(MSG_CHKSUM_ERROR);
					}
					else
					{
						chksum=(page>>8)%256;			//Checksum auch ber die commanddaten bilden
						chksum=(chksum+(page&0xff))%256;
						chksum=(chksum+(size>>8))%256;
						chksum=(chksum+(size&0xff))%256;
						msgpos=0;
						do
						{
							c=recchar();
							message[msgpos++]=c;
							chksum=(chksum+c)%256;
						} while(msgpos<size);
						if(size%2)
							message[msgpos++]=0xff;		//fllbyte einfgen wenn size nicht durch 2 teilbar
						c=recchar();
						if(!(page<APP_END))
						{
							//Fehler: wir wollen uns nicht selbst berschreiben
							sendchar(MSG_OVERWRITE_ERROR);
						}
						else
						{
							if(chksum!=c)
							{
								//Fehler: block nicht sauber bertragen
								sendchar(MSG_CHKSUM_ERROR);
							}
							else
							{
								uint8_t sreg = SREG;
								unsigned char *p = message;
								cli();
								for(i=0; i<size; i+=2)
								{
									// Set up little-endian word.
									uint16_t w = *p++;
									w += (*p++) << 8;
									boot_page_fill(page + i, w);
								}
								boot_page_write_safe(page);     // Store buffer in flash page.
								// Reenable RWW-section again. We need this if we want to jump back
								// to the application after bootloading.
								boot_rww_enable_safe();
								SREG=sreg;		//reenable interrupts (if they were enabled)
							}
						}
					}
					sendchar(MSG_END);
					PORTLED2 ^= (1<<PINLED2);
					break;
				case CMD_READ_FLASH:					//Command read_flash
					page=(recchar()<<8)|recchar();
					size=(recchar()<<8)|recchar();
					chksum=0;
					for(msgpos=0;msgpos<size;msgpos++)
					{
						c=pgm_read_byte_near(page+msgpos);
						sendchar(c);
						chksum=(chksum+c)%256;
					}
					sendchar(chksum);
					sendchar(MSG_END);
					PORTLED2 ^= (1<<PINLED2);
					break;
				case CMD_FLASH_BLOCKSIZE:				//Command get_pagesize
					sendchar(SPM_PAGESIZE>>8);
					sendchar(SPM_PAGESIZE&0xff);
					sendchar(MSG_END);
					break;
				case CMD_EEPROM_BLOCKSIZE:				//Command get_pagesize
					sendchar(E2END>>8);
					sendchar(E2END&0xff);
					sendchar(MSG_END);
					break;
				case CMD_WRITE_EEPROM:					//Command write eeprom
					page=(recchar()<<8)|recchar();
					size=(recchar()<<8)|recchar();
					if(!((page+size)<E2END))
					{
						//Fehler: block grer als eeprom
						sendchar(MSG_CHKSUM_ERROR);
					}
					else
					{
						chksum=(page>>8)%256;			//Checksum auch ber die commanddaten bilden
						chksum=(chksum+(page&0xff))%256;
						chksum=(chksum+(size>>8))%256;
						chksum=(chksum+(size&0xff))%256;
						msgpos=0;
						do
						{
							c=recchar();
							message[msgpos++]=c;
							chksum=(chksum+c)%256;
						} while(msgpos<size);
						c=recchar();
						if(chksum!=c)
						{
							//Fehler: block nicht sauber bertragen
							sendchar(MSG_CHKSUM_ERROR);
						}
						else
						{
							uint8_t sreg = SREG;
							cli();
							for(msgpos=0;msgpos<size;msgpos++)
							{
								EEAR=page+msgpos;
								EEDR=message[msgpos];
								EECR|=(1<<EEMWE);		//write eeprom
								EECR|=(1<<EEWE);
								while(EECR&(1<<EEWE))		//auf ende der schreiboperation warten
									;
							}
							SREG=sreg;
						}
					}
					sendchar(MSG_END);
					PORTLED2 ^= (1<<PINLED2);
					break;
				case CMD_READ_EEPROM:					//Command read eeprom
					page=(recchar()<<8)|recchar();
					size=(recchar()<<8)|recchar();
					if(!((page+size)<E2END))
					{
						//Fehler: block grer als eeprom
						sendchar(MSG_OVERWRITE_ERROR);
					}
					else
					{
						chksum=0;
						for(msgpos=0;msgpos<size;msgpos++)
						{
							EEAR=page+msgpos;
							EECR|=(1<<EERE);			//read eeprom
							c=EEDR;
							sendchar(c);
							chksum=(chksum+c)%256;
						}
						sendchar(chksum);
					}
					sendchar(MSG_END);
					PORTLED2 ^= (1<<PINLED2);
					break;
				case CMD_READ_LFUSE:					//Command read low fuse byte
					sendchar(boot_lock_fuse_bits_get( GET_LOW_FUSE_BITS ));
					sendchar(MSG_END);
					break;
				case CMD_READ_HFUSE:					//Command read high fuse byte
					sendchar(boot_lock_fuse_bits_get( GET_HIGH_FUSE_BITS ));
					sendchar(MSG_END);
					break;
				case CMD_JUMP_PROGRAM:					//Command enter_program_code
					sendchar(MSG_END);
					stop=0;
					break;
			}
		}//for
	}
	DDRLED1 = 0;	// set to default     
	DDRLED2 = 0;
	PORTLED1 = 0;
	PORTLED2 = 0;
	UART_CONTROL_REG &= ~((1 << UART_ENABLE_RECEIVER) | (1 << UART_ENABLE_TRANSMITTER));
	//wait 1 second
	for(c=0; c<100; c++)
		_delay_ms(10);
	asm volatile ( "push r1" "\n\t"  // Jump to Reset vector in Application Section
			"push r1" "\n\t" 
			"ret"     "\n\t" 
	::); 
	
	/*
	* Never return to stop GCC to generate exit return code
	* Actually we will never reach this point, but the compiler doesn't 
	* understand this
	*/
	for(;;);
}

